//go:build linux

package main

// Access to the kernel keyring for storing the unlocking pin securely in the
// user session. Copied and reduced from github.com/99designs/keyring (MIT).
//
// The original file is generalized and a better reference to use. An objective
// of rook is to minimize the amount of code to restrict the scope for errors;
// this file has been trimmed down and made specific to rook.

import (
	"errors"
	"syscall"

	"golang.org/x/sys/unix"
)

//nolint:revive
const (
	USER = "user"
)

// ROOKPIN is a variable so we can test it without interfering with a running rook
var ROOKPIN = "rookpin"

func getPin() ([]byte, error) {
	keyID, err := searchPin()
	if err != nil {
		return nil, err
	}

	var buffer []byte

	for {
		length, err := unix.KeyctlBuffer(unix.KEYCTL_READ, int(keyID), buffer, 0)
		if err != nil {
			return nil, err
		}

		// Return the buffer if it was large enough
		if length <= len(buffer) {
			return buffer[:length], nil
		}

		// Next try with a larger buffer
		buffer = make([]byte, length)
	}
}

func setPin(pin []byte) error {
	// If keysearch, delete key first, because Go's interface doesn't support update
	err := removePin()
	// If the key doesn't exist, that's fine; otherwise, it isn't.
	if err != nil && !errors.Is(err, ErrKeyNotFound{}) {
		return err
	}
	_, err = unix.AddKey(USER, ROOKPIN, pin, unix.KEY_SPEC_SESSION_KEYRING)
	if err != nil {
		return err
	}
	return nil
}

func removePin() error {
	keyID, err := searchPin()
	// If we can't find the key, we return ErrKeyNotFound(); the caller can decide whether that's
	// an error
	if err != nil {
		return err
	}
	parent := int32(unix.KEY_SPEC_SESSION_KEYRING)
	_, _, errno := syscall.Syscall(syscall.SYS_KEYCTL, uintptr(unix.KEYCTL_UNLINK), uintptr(keyID), uintptr(parent))
	// The key exists, but we can't delete it, which is a problem that we can't swallow
	if errno != 0 {
		return errno
	}
	return err
}

func searchPin() (keyID int, err error) {
	keyID, err = unix.KeyctlSearch(unix.KEY_SPEC_SESSION_KEYRING, USER, ROOKPIN, 0)
	if err != nil {
		if errors.Is(err, syscall.ENOKEY) {
			return -1, ErrKeyNotFound{}
		}
		return -1, err
	}
	return keyID, err
}
